//Ethan Alexander Shulman 2022 "Gilded Obsidian Fortress", February 25, 2022

//scene rendering options
#define VIEW_STEPS 80
#define VIEW_STEP_DITHER .3
#define VIEW_EPSILON 1e-4
#define VIEW_EPSILON_MUL 1.2
#define LIGHTING_STEPS 80
//#define LIGHTING_STEP_DITHER .3
#define LIGHTING_EPSILON 1e-2
//#define COLOR_MAX 100.
#define DENOISE_RANGE 10
#define DENOISE_NORMAL_THRESHOLD .9
#define DENOISE_DEPTH_THRESHOLD 1e1
#define DENOISE_HASH_DIVIDE 1
#define ANIMATION_TIME 10.
#define ANIMATION_SAMPLES 1000
#ifdef ANIMATION_TIME
out float time;
#endif

//constants
#define PI 3.14159265





mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))



/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	return mat3(21.3,-14.9+time,17.5,
			0,.5-time/10.,0,
			.7,0.,0.);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(mt,rc,gc,bc) if (chromaFrame < 1.) {mt[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mt[0]*=vec3(0,3,0);gc;}\
else {mt[0]*=vec3(0,0,3);bc;}



//functions defining scene
float sbox(vec3 p, vec3 sz) {
	p = abs(p)-sz;
	return length(max(p,0.))+min(max(p.x,max(p.y,p.z)),0.);
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//structure
	vec2 geo;
	if (p.y < 0.) {
		geo = vec2(-1e20,0);
		float fys = clamp(p.y+10.,0,1)*.2;
		for (float f = .5; f < 16.; f += f) {
			vec3 rp = mod(abs(p),f)-f*.5;
			geo.x = max(geo.x,-sbox(rp,vec3(f*.33)*vec3(1,1.+f*fys,1)));
		}
		geo.x = max(p.y,geo.x);
	} else {
		geo = vec2(p.y,0);
	}
	
	//fog
	if (p.y < 0. && p.y > -8. && geo.x > .4) {
		START_VOLUME(p);
		float fdst = VOLUME(p.y,.4,.2,1.);
		if (fdst < geo.x) geo = vec2(fdst,1.);
	}
	
	return geo;
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, float mat) {
	if (mat == 1.) {
		//fog
		return mat3(vec3(1), .5,0,1, 1,0,0);
	}

	//structure
	//gold gilding
	if (mod(abs(p.x+.08),2.) < .2) {
		return mat3(vec3(.8,.65,.24), .0014,0,1, 1,0,0);
	}
	//obsidian
	if (mod(abs(p.y+1.),4.) < 2.) {
		return mat3(vec3(.04), 0,0,1, 1,0,0);
	}
	
	//rough stone
	return mat3(vec3(.4), .6,0,1, 1,0,0);
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(.1,.1,.12)*max(0.,d.y)+max(0.,1.-2.*length(d-normalize(vec3(-3,1.5,1))))*1000.;
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return p.y>0.?background(p,d):vec3(0);
}

